# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/42_helper.date_and_time.ipynb.

# %% auto 0
__all__ = ['current_time_formatted_to_minutes']

# %% ../../nbs/42_helper.date_and_time.ipynb 2
import datetime
from datetime import timezone

# %% ../../nbs/42_helper.date_and_time.ipynb 4
def current_time_formatted_to_minutes(
        ) -> str:
    """Return the current time to minutes.

    **Returns**

    - str
        - In UTC time, to minutes.
    """
    dt = datetime.datetime.now(timezone.utc)
    formatted = dt.isoformat(timespec='minutes')
    return formatted[:16]
