"""A concise version of the code used in [tutorial.walkthrough](tutorial.walkthrough.html)"""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/tutorial.concise_code.ipynb.

# %% auto 0
__all__ = []

# %% ../../nbs/tutorial.concise_code.ipynb 3
import pathlib
from pathlib import Path
import platform
import warnings

from huggingface_hub import from_pretrained_fastai
from transformers import AutoModelForSeq2SeqLM, AutoModelForTokenClassification, AutoTokenizer, pipeline

from ..helper.arxiv import arxiv_id, arxiv_search, download_from_results, extract_last_names
from ..helper.files_and_folders import get_download_path, get_huggingface_cache_dir, text_from_file
from ..latex import find_main_latex_file
from trouver.latex.convert import (
    divide_preamble, divide_latex_text, custom_commands,
    setup_reference_from_latex_parts
)
from ..latex.preamble import replace_inclusion_of_style_file_with_code
from ..markdown.markdown.file import MarkdownFile
from ..markdown.obsidian.personal.machine_learning.information_note_types import automatically_add_note_type_tags
from ..markdown.obsidian.personal.machine_learning.tokenize.def_and_notat_token_classification import auto_mark_def_and_notats
from ..markdown.obsidian.personal.machine_learning.definition_and_notation_naming import add_names_to_html_tags_in_info_note
from ..markdown.obsidian.personal.machine_learning.notation_summarization import append_summary_to_notation_note
from ..markdown.obsidian.personal.notes import notes_linked_in_notes_linked_in_note, notes_linked_in_note
from ..markdown.obsidian.personal.notation import make_notation_notes_from_double_asts, make_notation_notes_from_HTML_tags
from ..markdown.obsidian.personal.notation.in_standard_information_note import notation_notes_linked_in_see_also_section
from ..markdown.obsidian.vault import VaultNote
