"""Functions relevant to gathering and processing machine learning data."""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../../../../../nbs/17_markdown.obsidian.personal.machine_learning.__init__.ipynb.

# %% auto 0
__all__ = []

# %% ../../../../../nbs/17_markdown.obsidian.personal.machine_learning.__init__.ipynb 2
# from huggingface_hub import HfApi, hf_hub_download

# %% ../../../../../nbs/17_markdown.obsidian.personal.machine_learning.__init__.ipynb 3
# def get_model_size(model_name) -> int:
#     """
#     Return the size of the model in bytes.
#     """
#     # Get model info
#     api = HfApi()
#     model_info = api.model_info(model_name)
    
#     # Calculate total size
#     total_size = sum(file.size for file in model_info.siblings)
#     return total_size
