"""Functions relevant for training models to identify notations introduced in mathematical text."""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../../../../../nbs/18_markdown.obsidian.personal.machine_learning.notations.ipynb.

# %% auto 0
__all__ = []

# %% ../../../../../nbs/18_markdown.obsidian.personal.machine_learning.notations.ipynb 2
#| export
