"""Functions for gathering data to train and for using tokenization machine learning models."""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../../../../../../nbs/26_markdown.obsidian.personal.machine_learning.tokenize.__init__.ipynb.

# %% auto 0
__all__ = []

# %% ../../../../../../nbs/26_markdown.obsidian.personal.machine_learning.tokenize.__init__.ipynb 2
#| export
