"""Functions for gathering and processing data on determining where to divide mathematical text"""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../../../../../nbs/53_markdown.obsidian.personal.machine_learning.note_division.ipynb.

# %% auto 0
__all__ = ['combine_info_notes_in_index_notes']

# %% ../../../../../nbs/53_markdown.obsidian.personal.machine_learning.note_division.ipynb 3
from ...vault import VaultNote

# %% ../../../../../nbs/53_markdown.obsidian.personal.machine_learning.note_division.ipynb 5
def combine_info_notes_in_index_notes(
        index_note: VaultNote,
        ) -> tuple[str, list[int]]: 
    """
    
    """
    return
